/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*- */

/*
 * Copyright (C) 2002 Thomas Vander Stichele
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Author: Thomas Vander Stichele <thomas at apestaart dot org>
 */

/* audio-play.c - play files using AudioPlay GObject
 */

#include <gst/gst.h>
#include "audio-play.h"

#define DEBUG
/* #define DEBUG_XML */
#ifdef DEBUG_XML
#include <stdio.h>
FILE *_debug_file;
#endif

/* private object properties */
struct AudioPlayPriv {
	GstElement *pipeline;		/* main playback pipeline */
	GstElement *src;
	GstElement *typefind;
	GstElement *decoder;
	GstElement *sink;

	GstCaps *caps;			/* FIXME: why ? */

	GstClock *clock;
	gboolean have_type;
	gboolean have_eos;		/* dispatcher booleans from idler */
	gboolean have_cache_empty;
	gboolean have_caps;

	gchar *location;

	guint64 length_nanoseconds;	/* total length */
	gint seconds;			/* current playing time */
	guint64 nanoseconds;

	guint idler_id;
	guint iterator_id;
};

/* signal enum */
enum {
	EOS_SIGNAL,
	TICK_SIGNAL,
	LENGTH_SIGNAL,
	LAST_SIGNAL
};

/* GError enum */
enum {
	AUDIO_PLAY_ERROR_AUTOPLUG
};

static guint audio_play_signals [LAST_SIGNAL] = { 0 };
	

/* GError quark stuff */
GQuark
audio_play_error_quark (void)
{
	static GQuark quark = 0;
	if (quark == 0) 
		quark = g_quark_from_static_string ("audio-play-error-quark");
	return quark;
}

/* forward declarations */
static void	audio_play_class_init 	(AudioPlayClass *klass);
static void	audio_play_init 	(AudioPlay *play);

static gboolean	audio_play_playing_to_ready (AudioPlay *play);
//static void	gst_play_dispose	(GObject *object);

/* GObject type stuff */

static GstElementClass *parent_class = NULL;

GType
audio_play_get_type (void)
{
	static GType play_type = 0;

	if (!play_type)
	{
		static const GTypeInfo play_info = {
			sizeof (AudioPlayClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) audio_play_class_init,
			NULL, NULL, sizeof (AudioPlay),
			0, (GInstanceInitFunc) audio_play_init,
			NULL
		};
		play_type = g_type_register_static (G_TYPE_OBJECT, "AudioPlay", 
			                    	    &play_info, 0);
										        }
	return play_type;
}


/* private functions */
/* the lazy man's version of debugging */
void
audio_play_debug (AudioPlay *play)
{
	g_print ("DEBUG: audio_play %p\n", play);
	g_print ("DEBUG: pipeline %p\n", play->priv->pipeline);
	g_print ("DEBUG: typefind %p\n", play->priv->typefind);
	g_print ("DEBUG: decoder %p\n", play->priv->decoder);
	g_print ("DEBUG: time %d\n", play->priv->seconds);
}


/* callbacks & idlers */
/* iterator idler func */
gboolean
iterator (AudioPlay *play)
{
#ifdef DEBUG
	g_print ("+");
#endif
	if (gst_bin_iterate (GST_BIN (play->priv->pipeline)))
		return TRUE;
	else
	{
		g_print ("DEBUG: iterator: couldn't iterate\n");
		/* do an eos */
		play->priv->have_eos = TRUE;
		return FALSE;
	}
}

/* main idler func, handles all sorts of requests asynchronously */
gboolean
idler (AudioPlay *play)
{
#ifdef DEBUG
	g_print ("-");
#endif
	if (play->priv->have_eos)
	{
		g_print ("DEBUG: signalling eos\n");
		play->priv->have_eos = FALSE;
		audio_play_playing_to_ready (play);
		/* gst_main_quit (); */
		g_signal_emit (G_OBJECT (play), 
			       audio_play_signals [EOS_SIGNAL], 
			       0);
		return FALSE;
	}
	if (audio_play_get_state (play) == GST_STATE_PLAYING)
		return TRUE;
	else
	{
		g_print ("DEBUG: idler returning false, not playing\n");
		//return FALSE;
	}
}

/* timeout func for emitting the tick signal */
static gboolean
tick_timeout (AudioPlay *play)
{
	gint secs;
	guint64 nanosecs;
	GstClock *clock = play->priv->clock;
	AudioPlayPriv *priv = play->priv;

	nanosecs = gst_clock_get_time (clock);
	secs = (gint) (nanosecs / GST_SECOND);
	if (secs != priv->seconds)
	{
		priv->seconds = secs;
		priv->nanoseconds = nanosecs;
		g_signal_emit (G_OBJECT (play), 
			       audio_play_signals [TICK_SIGNAL], 
			       0, priv->nanoseconds);
	}
	/* remove ourselves when we stop playing */
	return (gst_element_get_state (priv->pipeline) == GST_STATE_PLAYING);
}


void
callback_sink_eos (GstElement *element, AudioPlay *play)
{
	AudioPlayPriv *priv = play->priv;

  	g_print ("DEBUG: sink_eos triggered\n");
	play->priv->have_eos = TRUE;
	/* gst_main_quit (); */
}

/* callback used by query of length of track to know when caps are there */
static void
deep_notify (GObject *object, GstObject *origin,
	     GParamSpec *pspec, AudioPlay *play)
{
	GValue value = { 0 };
	/* don't care about file offsets */
	if (strcmp (pspec->name, "offset") == 0) return;
	if (strcmp (pspec->name, "caps") == 0)
	{
		g_print ("DEBUG: got caps\n");
		g_value_init (&value, pspec->value_type);
		g_object_get_property (G_OBJECT (origin), pspec->name, &value);
		play->priv->caps = g_value_peek_pointer (&value);
		play->priv->have_caps = TRUE;
	}
	if (strcmp (pspec->name, "channels") == 0)
	{
		/* good enough to have_caps on */
		play->priv->have_caps = TRUE;
	}
	g_print ("DEBUG: deep_notify: %s\n", pspec->name);
}
/* callback for when we have the type of the file 
 * we set up a decoder based on the mime type
 */
static void 
have_type (GstElement *element, GstCaps *caps, AudioPlay *play) 
{
	AudioPlayPriv *priv = play->priv;
  	gchar *mime = g_strdup_printf (gst_caps_get_mime (caps));
	GstElement *decoder = NULL;

#ifdef DEBUG
	g_print ("DEBUG: have_type callback called.\n");
  	fprintf (stderr, "have caps, mime type is %s\n", mime);
#endif

	g_assert (GST_IS_ELEMENT (priv->pipeline));
  	gst_element_set_state (priv->pipeline, GST_STATE_READY);

  	/* disconnect the typefind from the pipeline and remove it,
   	 * since we now know what the type is */
	gst_element_disconnect (priv->src, priv->typefind);
	/* ref it so it doesn't go away */
	g_object_ref (priv->typefind);
  	gst_bin_remove (GST_BIN (priv->pipeline), priv->typefind);

  	/* gst_scheduler_show (GST_ELEMENT_SCHED (priv->pipeline)); */

  	/* now based on the mime type set up the pipeline properly */
  	if (strstr (mime, "mp3")) 
    		decoder = gst_element_factory_make ("mad", "decoder");
  	else if (strstr (mime, "x-ogg"))
    		decoder = gst_element_factory_make ("vorbisfile", "decoder");
  	else if (strstr (mime, "x-wav"))
    		decoder = gst_element_factory_make ("wavparse", "decoder");
  	else if (strstr (mime, "x-flac"))
    		decoder = gst_element_factory_make ("flacdec", "decoder");
  	else 
	{
    		g_print ("mime type %s not handled in this program.\n", mime);
    		g_free (mime);
		return;
	}
  	g_free (mime);
	priv->decoder = decoder;

	/* set it to null */
  	gst_element_set_state (priv->pipeline, GST_STATE_NULL);

  	/* now put playback pipe in place of where typefind was  */
	g_print ("DEBUG: putting in decoder\n");
  	gst_bin_add (GST_BIN (priv->pipeline), priv->decoder);
	g_print ("DEBUG: putting in sink\n");
  	gst_bin_add (GST_BIN (priv->pipeline), priv->sink);
	/* FIXME: this shouldn't go here */
	g_signal_connect (G_OBJECT (play->priv->sink), "eos",
			  G_CALLBACK (callback_sink_eos), play);
	if (!gst_element_connect (priv->src, priv->decoder))
		g_error ("warning: could not connect src and decoder !\n");
	if (!gst_element_connect (priv->decoder, priv->sink))
		g_print ("warning: could not connect decoder and sink !\n");
  	if (gst_element_set_state (priv->pipeline, GST_STATE_PAUSED)
			== GST_STATE_FAILURE)
		g_print ("pipelien could not be paused !\n");
  	/* gst_scheduler_show (GST_ELEMENT_SCHED (priv->pipeline)); */
}
/* object functions */

static void
audio_play_class_init (AudioPlayClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

        parent_class = g_type_class_ref (GST_TYPE_OBJECT);

        klass->eos = NULL;

        //gobject_class->dispose = audio_play_dispose;

        audio_play_signals [EOS_SIGNAL] =
		g_signal_new ("eos",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_FIRST,
                              G_STRUCT_OFFSET (AudioPlayClass, eos),
                              NULL, NULL,
                              gst_marshal_VOID__VOID,
                              G_TYPE_NONE, 0);
        audio_play_signals [LENGTH_SIGNAL] =
		g_signal_new ("length",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_FIRST,
                              G_STRUCT_OFFSET (AudioPlayClass, length),
                              NULL, NULL,
                              gst_marshal_VOID__INT64,
                              G_TYPE_NONE, 1,
			      G_TYPE_INT64);
        audio_play_signals [TICK_SIGNAL] =
		g_signal_new ("tick",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_FIRST,
                              G_STRUCT_OFFSET (AudioPlayClass, tick),
                              NULL, NULL,
                              gst_marshal_VOID__INT64,
                              G_TYPE_NONE, 1,
			      G_TYPE_INT64);
}

static void
audio_play_init (AudioPlay *play)
{
	AudioPlayPriv *priv = g_malloc (sizeof (AudioPlayPriv));

	priv->pipeline = NULL;
	priv->src = NULL;
	priv->typefind = NULL;
	priv->decoder = NULL;
	priv->sink = NULL;

	priv->caps = NULL;
	priv->seconds = 0;
	priv->nanoseconds = 0;
	priv->length_nanoseconds = 0;

	priv->location = NULL;

	priv->idler_id = 0;
	priv->iterator_id = 0;

	priv->have_caps = FALSE;

	play->priv = priv;
}

/* create audio play object */
/* makes a pipeline with gnomevfsssrc in it
 * and pre-initializes typefind
 */

AudioPlay * 
audio_play_new (GError **error)
{
	AudioPlay *play = g_object_new (AUDIO_TYPE_PLAY, NULL);
	AudioPlayPriv *priv = play->priv;

	priv->pipeline = gst_pipeline_new ("pipeline");
	g_assert (priv->pipeline);

	/* set up all the elements we're going to use */

	priv->src = gst_element_factory_make ("gnomevfssrc", "src");
	g_assert (priv->src);

	priv->typefind = gst_element_factory_make ("typefind", "typefind");
	g_assert (priv->typefind);

	gst_bin_add (GST_BIN (priv->pipeline), priv->src);

	/* execute the callback when we find the type */
	g_signal_connect (priv->typefind, "have_type",
		          G_CALLBACK (have_type), play);

	/* catch deep notifies */
	g_signal_connect (priv->pipeline, "deep_notify",
		          G_CALLBACK (deep_notify), play);
	priv->have_type = FALSE;
	priv->have_eos = FALSE;
	play->priv = priv;

	g_print ("DEBUG: audio_play_new done\n");
	audio_play_debug (play);
	return play;
}

/* free all resources held by the play object */
void
audio_play_dispose (AudioPlay *play)
{
	AudioPlayPriv *priv;
       
	g_print ("DEBUG: audio_play_dispose\n");
	g_assert (play != NULL);
	g_return_if_fail (IS_AUDIO_PLAY (play));
	if (gst_element_get_state (play->priv->pipeline) == GST_STATE_PLAYING)
		audio_play_playing_to_ready (play);	

	g_print ("DEBUG: dispose: getting priv\n");
	priv = play->priv;

	gst_element_set_state (priv->pipeline, GST_STATE_NULL);
	g_print ("DEBUG: dispose: set to null\n");
	/*
	if (priv->decoder)
	{
		gst_bin_remove (GST_BIN (priv->pipeline), priv->decoder);
	}
	g_object_unref (priv->typefind);
	g_object_unref (priv->src);
	g_object_unref (priv->sink);
	g_object_unref (priv->pipeline);
	*/
}

/* prerequisites:
 * - no decoder in pipeline
 * - typefind exists but not yet in pipeline
 */
static gboolean
audio_play_autoplug (AudioPlay *play, GError **error)
{
	AudioPlayPriv *priv = play->priv;
#ifdef DEBUG
	g_print ("DEBUG: starting autoplug\n");
#endif
	g_assert (GST_IS_BIN (priv->pipeline));

	if (gst_element_get_state (priv->pipeline) != GST_STATE_READY)
	{
		g_print ("Can't autoplug, not in READY state\n");
		return FALSE;
	}
	g_assert (play->priv->decoder == NULL);

	/* put in typefind */
	g_print ("putting in typefind\n");
	gst_bin_add (GST_BIN (priv->pipeline), priv->typefind);
	gst_element_connect (priv->src, priv->typefind);

	/* iterate pipe until type is found through callback */
	gst_element_set_state ((play->priv->pipeline), GST_STATE_PLAYING);
	/* gst_scheduler_show (GST_ELEMENT_SCHED (play->priv->pipeline)); */
	audio_play_debug (play);
	g_print ("DEBUG: starting iterating for autoplug\n");
	g_assert (gst_element_get_state (play->priv->pipeline) == GST_STATE_PLAYING);
	while (play->priv->decoder == NULL)
	{
		g_print ("*");
	       if (!gst_bin_iterate (GST_BIN (play->priv->pipeline)))
	       {
		       /* clean up typefind */
			gst_element_disconnect (priv->src, priv->typefind);
			g_object_ref (priv->typefind);
  			gst_bin_remove (GST_BIN (priv->pipeline), 
					priv->typefind);
		       if (!error) return FALSE;
		       *error = g_error_new (AUDIO_PLAY_ERROR,
				             AUDIO_PLAY_ERROR_AUTOPLUG,
					     "Could not find a codec for %s\n",
					     play->priv->location);
		       return FALSE;
	       }
	}
	/* we have the type */
#ifdef DEBUG
	g_print ("DEBUG: done autoplug\n");
#endif
	return TRUE;
}

/* load the location into the play object */
gboolean
audio_play_set_location (AudioPlay *play, const char *uri, GError **error)
{
#ifdef DEBUG
	g_print ("DEBUG: loading location %s\n", uri);
#endif
	if (gst_element_get_state (play->priv->pipeline) == GST_STATE_PLAYING)
	{
		audio_play_playing_to_ready (play);
	}
	gst_element_set_state ((play->priv->pipeline), GST_STATE_READY);
	if (play->priv->location) g_free (play->priv->location);
	play->priv->location = g_strdup (uri);
	g_object_set (G_OBJECT (play->priv->src), "location", uri, NULL);

	/* clear values that need clearing */
	play->priv->seconds = 0;
	play->priv->nanoseconds = 0;
	play->priv->length_nanoseconds = 0;
	play->priv->have_caps = FALSE;

	/* do a typefind */
	audio_play_autoplug (play, error);
}


/* (re)query length of track */
static void
audio_play_update_length (AudioPlay *play)
{
	gboolean res;
	GstFormat format = GST_FORMAT_TIME;
	GstPad *pad;
	gint64 value;
	GstCaps *caps;
	const GstFormatDefinition *definition = gst_format_get_details (format);

	g_assert (IS_AUDIO_PLAY (play));
	
	while (gst_bin_iterate (GST_BIN (play->priv->pipeline)) &&
	       play->priv->have_caps == FALSE)
		g_print ("?");
	g_print ("\n");

	pad = gst_element_get_pad (play->priv->decoder, "src");
	if (pad == NULL)
	{
		g_print ("WARNING: could not get sink pad of decoder !\n");
		return;
	}
	caps = gst_pad_get_caps (pad);
	if (caps == NULL) g_print ("WARNING: caps are NULL\n");
	gst_caps_debug (caps, "caps of decoder src pad");

	res = gst_pad_query (pad, GST_PAD_QUERY_TOTAL, &format, &value);
	if (!res) 
	{
		g_print ("WARNING: pad query failed,can't get length !\n");
		play->priv->length_nanoseconds = FALSE;
		return;
	}
	g_assert (res == TRUE);
	g_assert (format == GST_FORMAT_TIME);
	g_print ("DEBUG: length in %s: %lld (nanosecs)\n", definition->nick, 
		 value);
	play->priv->length_nanoseconds = value;
	g_signal_emit (G_OBJECT (play), 
	               audio_play_signals [LENGTH_SIGNAL], 0,
		       play->priv->length_nanoseconds);
}

/* seek to given time (in nanoseconds) */
void
audio_play_seek_to_time (AudioPlay *play, guint64 nanosecs)
{
	GstEvent *event;
	gint64 seek_time;

	if (nanosecs < 0LL)
		seek_time = 0LL;
	else
		seek_time = (gint64) nanosecs;
	g_print ("DEBUG: audio_play_seek_to_time: seeking to %lld\n", seek_time);
	gst_element_set_state (play->priv->pipeline, GST_STATE_PAUSED);
	event = gst_event_new_seek (GST_FORMAT_TIME |
			            GST_SEEK_METHOD_SET |
				    GST_SEEK_FLAG_FLUSH, seek_time);
	/* FIXME: do we need to ref before sending ? */
	gst_event_ref (event);
	if (gst_element_send_event (play->priv->sink, event))
	{
		play->priv->nanoseconds = nanosecs;
		g_signal_emit (G_OBJECT (play), 
			       audio_play_signals [TICK_SIGNAL], 0,
			       play->priv->nanoseconds);
	}
	gst_element_set_state (play->priv->pipeline, GST_STATE_PLAYING);
}

/* seek to given position (0.0 - 1.0) */
void
audio_play_seek_to_pos (AudioPlay *play, double value)
{
	guint64 seek_to = (guint64) ((double) play->priv->length_nanoseconds * value);
	g_print ("DEBUG: audio_play_seek_to_pos: seeking to %lld\n", seek_to);
	g_print ("DEBUG: value was %f\n", value);
	audio_play_seek_to_time (play, seek_to);
}

gboolean
audio_play_set_state (AudioPlay *play, GstElementState state, GError **error)
{
	switch (state)
	{
		case GST_STATE_PLAYING:
			g_print ("DEBUG: playing audio play\n");
			/* FIXME: somewhere else ? update clock */
			play->priv->clock = 
				gst_bin_get_clock (GST_BIN (play->priv->pipeline));
			if (play->priv->decoder == NULL)
			{
				g_print ("DEBUG: can't play, no decoder yet\n");
				return FALSE;
			}
			gst_element_set_state (play->priv->pipeline, 
					       GST_STATE_PLAYING);
			audio_play_update_length (play);

			/* add idlers */
			/* handle signals asyncly */
			play->priv->idler_id = 
				g_idle_add ((GSourceFunc) idler, play);
			g_assert (play->priv->idler_id > 0);
			/* set handler for playback; it will unregister
			 * when iterate fails */
			play->priv->iterator_id = 
				g_idle_add ((GSourceFunc) iterator, 
				            play);
			g_assert (play->priv->iterator_id > 0);
			/* set tick timeout */
			g_timeout_add (200, (GSourceFunc) tick_timeout, play);
			break;
		case GST_STATE_PAUSED:
			g_print ("DEBUG: pausing audio play\n");
			gst_element_set_state (play->priv->pipeline,
					       GST_STATE_PAUSED);
			break;
		case GST_STATE_READY:
			g_print ("DEBUG: going to ready");
			if (play->priv->decoder)
			{
				if (!audio_play_playing_to_ready (play))
					g_print ("couldn't play->ready\n");
			}
			g_print ("DEBUG: decoder: %p\n", play->priv->decoder);
			break;
		default:
			g_print ("unhandled state\n");
			break;
	}
}

GstElementState
audio_play_get_state (AudioPlay *play)
{
	return gst_element_get_state (play->priv->pipeline);
}

guint64
audio_play_get_length (AudioPlay *play)
{
	g_print ("DEBUG: audio_play_get_length: returning %lld\n",
		 play->priv->length_nanoseconds);
	return play->priv->length_nanoseconds;
}

/* set an audio sink on the play object to use.  It must be initialized */
void
audio_play_set_audio_sink (AudioPlay *play, GstElement *element)
{
	g_assert (GST_IS_ELEMENT (element));
	if (play->priv->sink != NULL)
	{
		g_warning ("Already have a sink !\n");
		return;
	}
	play->priv->sink = element;
	g_signal_connect (G_OBJECT (play->priv->sink), "eos",
			  G_CALLBACK (callback_sink_eos), play);
}

/* move from playing to ready */
static gboolean
audio_play_playing_to_ready (AudioPlay *play)
{
	AudioPlayPriv *priv = play->priv;
	/* stop the pipepline */
	gst_element_set_state (priv->pipeline, GST_STATE_READY);
	
	/* disconnect src ! decoder ! sink */
	gst_element_disconnect (priv->src, priv->decoder);
	gst_element_disconnect (priv->decoder, priv->sink);
	/* remove sink and decoder from the bin */
	gst_bin_remove (GST_BIN (priv->pipeline), priv->decoder);
	g_object_ref (priv->sink);
	gst_bin_remove (GST_BIN (priv->pipeline), priv->sink);

	/* null out decoder */
	priv->decoder = NULL;

	/* if idle functions are still going, remove them */
	if (priv->iterator_id != 0)
	{
		g_source_remove (priv->iterator_id);
		priv->iterator_id = 0;
	}
	if (priv->idler_id != 0)
	{
		g_source_remove (priv->idler_id);
		priv->idler_id = 0;
	}


	/* reconnect typefind to source and readd to bin */
	/*
	g_print ("DEBUG: reconnecting typefind\n");
	gst_element_connect (priv->src, priv->typefind);
	gst_bin_add (GST_BIN (priv->pipeline), priv->typefind);
	*/
	g_print ("DEBUG: play to ready done\n");
	return TRUE;
}
